@@ -55,11 +55,11 @@ args_cmp(struct args_entry *a1, struct args_entry *a2)
 
 /* Find a flag in the arguments tree. */
 static struct args_entry *
-args_find(struct args *args, u_char ch)
+args_find(struct args *args, u_char flag)
 {
 	struct args_entry	entry;
 
-	entry.flag = ch;
+	entry.flag = flag;
 	return (RB_FIND(args_tree, &args->tree, &entry));
 }
 
@@ -74,6 +74,7 @@ args_parse(const char *template, int argc, char **argv)
 
 	optreset = 1;
 	optind = 1;
+	optarg = NULL;
 
 	while ((opt = getopt(argc, argv, template)) != -1) {
 		if (opt < 0)
@@ -83,6 +84,7 @@ args_parse(const char *template, int argc, char **argv)
 			return (NULL);
 		}
 		args_set(args, opt, optarg);
+		optarg = NULL;
 	}
 	argc -= optind;
 	argv += optind;
@@ -209,25 +211,35 @@ args_print(struct args *args)
 char *
 args_escape(const char *s)
 {
-	static const char	 quoted[] = " #\"';${}";
+	static const char	 dquoted[] = " #';${}";
+	static const char	 squoted[] = " \"";
 	char			*escaped, *result;
-	int			 flags;
+	int			 flags, quotes = 0;
+
+	if (*s == '\0') {
+		xasprintf(&result, "''");
+		return (result);
+	}
+	if (s[strcspn(s, dquoted)] != '\0')
+		quotes = '"';
+	else if (s[strcspn(s, squoted)] != '\0')
+		quotes = '\'';
 
-	if (*s == '\0')
-		return (xstrdup(s));
 	if (s[0] != ' ' &&
-	    (strchr(quoted, s[0]) != NULL || s[0] == '~') &&
-	    s[1] == '\0') {
+	    s[1] == '\0' &&
+	    (quotes != 0 || s[0] == '~')) {
 		xasprintf(&escaped, "\\%c", s[0]);
 		return (escaped);
 	}
 
 	flags = VIS_OCTAL|VIS_CSTYLE|VIS_TAB|VIS_NL;
-	if (s[strcspn(s, quoted)] != '\0')
+	if (quotes == '"')
 		flags |= VIS_DQ;
 	utf8_stravis(&escaped, s, flags);
 
-	if (flags & VIS_DQ) {
+	if (quotes == '\'')
+		xasprintf(&result, "'%s'", escaped);
+	else if (quotes == '"') {
 		if (*escaped == '~')
 			xasprintf(&result, "\"\\%s\"", escaped);
 		else
@@ -244,11 +256,11 @@ args_escape(const char *s)
 
 /* Return if an argument is present. */
 int
-args_has(struct args *args, u_char ch)
+args_has(struct args *args, u_char flag)
 {
 	struct args_entry	*entry;
 
-	entry = args_find(args, ch);
+	entry = args_find(args, flag);
 	if (entry == NULL)
 		return (0);
 	return (entry->count);
@@ -256,15 +268,15 @@ args_has(struct args *args, u_char ch)
 
 /* Set argument value in the arguments tree. */
 void
-args_set(struct args *args, u_char ch, const char *s)
+args_set(struct args *args, u_char flag, const char *s)
 {
 	struct args_entry	*entry;
 	struct args_value	*value;
 
-	entry = args_find(args, ch);
+	entry = args_find(args, flag);
 	if (entry == NULL) {
 		entry = xcalloc(1, sizeof *entry);
-		entry->flag = ch;
+		entry->flag = flag;
 		entry->count = 1;
 		TAILQ_INIT(&entry->values);
 		RB_INSERT(args_tree, &args->tree, entry);
@@ -280,22 +292,44 @@ args_set(struct args *args, u_char ch, const char *s)
 
 /* Get argument value. Will be NULL if it isn't present. */
 const char *
-args_get(struct args *args, u_char ch)
+args_get(struct args *args, u_char flag)
 {
 	struct args_entry	*entry;
 
-	if ((entry = args_find(args, ch)) == NULL)
+	if ((entry = args_find(args, flag)) == NULL)
+		return (NULL);
+	if (TAILQ_EMPTY(&entry->values))
 		return (NULL);
 	return (TAILQ_LAST(&entry->values, args_values)->value);
 }
 
+/* Get first argument. */
+u_char
+args_first(struct args *args, struct args_entry **entry)
+{
+	*entry = RB_MIN(args_tree, &args->tree);
+	if (*entry == NULL)
+		return (0);
+	return ((*entry)->flag);
+}
+
+/* Get next argument. */
+u_char
+args_next(struct args_entry **entry)
+{
+	*entry = RB_NEXT(args_tree, &args->tree, *entry);
+	if (*entry == NULL)
+		return (0);
+	return ((*entry)->flag);
+}
+
 /* Get first value in argument. */
 const char *
-args_first_value(struct args *args, u_char ch, struct args_value **value)
+args_first_value(struct args *args, u_char flag, struct args_value **value)
 {
 	struct args_entry	*entry;
 
-	if ((entry = args_find(args, ch)) == NULL)
+	if ((entry = args_find(args, flag)) == NULL)
 		return (NULL);
 
 	*value = TAILQ_FIRST(&entry->values);
@@ -318,15 +352,15 @@ args_next_value(struct args_value **value)
 
 /* Convert an argument value to a number. */
 long long
-args_strtonum(struct args *args, u_char ch, long long minval, long long maxval,
-    char **cause)
+args_strtonum(struct args *args, u_char flag, long long minval,
+    long long maxval, char **cause)
 {
 	const char		*errstr;
 	long long 	 	 ll;
 	struct args_entry	*entry;
 	struct args_value	*value;
 
-	if ((entry = args_find(args, ch)) == NULL) {
+	if ((entry = args_find(args, flag)) == NULL) {
 		*cause = xstrdup("missing");
 		return (0);
 	}
@@ -341,3 +375,60 @@ args_strtonum(struct args *args, u_char ch, long long minval, long long maxval,
 	*cause = NULL;
 	return (ll);
 }
+
+/* Convert an argument to a number which may be a percentage. */
+long long
+args_percentage(struct args *args, u_char flag, long long minval,
+    long long maxval, long long curval, char **cause)
+{
+	const char		*value;
+	struct args_entry	*entry;
+
+	if ((entry = args_find(args, flag)) == NULL) {
+		*cause = xstrdup("missing");
+		return (0);
+	}
+	value = TAILQ_LAST(&entry->values, args_values)->value;
+	return (args_string_percentage(value, minval, maxval, curval, cause));
+}
+
+/* Convert a string to a number which may be a percentage. */
+long long
+args_string_percentage(const char *value, long long minval, long long maxval,
+    long long curval, char **cause)
+{
+	const char	*errstr;
+	long long 	 ll;
+	size_t		 valuelen = strlen(value);
+	char		*copy;
+
+	if (value[valuelen - 1] == '%') {
+		copy = xstrdup(value);
+		copy[valuelen - 1] = '\0';
+
+		ll = strtonum(copy, 0, 100, &errstr);
+		free(copy);
+		if (errstr != NULL) {
+			*cause = xstrdup(errstr);
+			return (0);
+		}
+		ll = (curval * ll) / 100;
+		if (ll < minval) {
+			*cause = xstrdup("too small");
+			return (0);
+		}
+		if (ll > maxval) {
+			*cause = xstrdup("too large");
+			return (0);
+		}
+	} else {
+		ll = strtonum(value, minval, maxval, &errstr);
+		if (errstr != NULL) {
+			*cause = xstrdup(errstr);
+			return (0);
+		}
+	}
+
+	*cause = NULL;
+	return (ll);
+}
