@@ -17,7 +17,9 @@
  */
 
 #include <sys/types.h>
+#include <sys/ioctl.h>
 #include <sys/socket.h>
+#include <sys/wait.h>
 
 #include <fcntl.h>
 #include <signal.h>
@@ -64,22 +66,20 @@ struct job {
 /* All jobs list. */
 static LIST_HEAD(joblist, job) all_jobs = LIST_HEAD_INITIALIZER(all_jobs);
 
-/* Start a job running, if it isn't already. */
+/* Start a job running. */
 struct job *
-job_run(const char *cmd, struct session *s, const char *cwd,
-    job_update_cb updatecb, job_complete_cb completecb, job_free_cb freecb,
-    void *data, int flags)
+job_run(const char *cmd, int argc, char **argv, struct session *s,
+    const char *cwd, job_update_cb updatecb, job_complete_cb completecb,
+    job_free_cb freecb, void *data, int flags, int sx, int sy)
 {
-	struct job	*job;
-	struct environ	*env;
-	pid_t		 pid;
-	int		 nullfd, out[2];
-	const char	*home;
-	sigset_t	 set, oldset;
-
-	if (socketpair(AF_UNIX, SOCK_STREAM, PF_UNSPEC, out) != 0)
-		return (NULL);
-	log_debug("%s: cmd=%s, cwd=%s", __func__, cmd, cwd == NULL ? "" : cwd);
+	struct job	 *job;
+	struct environ	 *env;
+	pid_t		  pid;
+	int		  nullfd, out[2], master;
+	const char	 *home;
+	sigset_t	  set, oldset;
+	struct winsize	  ws;
+	char		**argvp;
 
 	/*
 	 * Do not set TERM during .tmux.conf, it is nice to be able to use
@@ -89,56 +89,84 @@ job_run(const char *cmd, struct session *s, const char *cwd,
 
 	sigfillset(&set);
 	sigprocmask(SIG_BLOCK, &set, &oldset);
-	switch (pid = fork()) {
+
+	if (flags & JOB_PTY) {
+		memset(&ws, 0, sizeof ws);
+		ws.ws_col = sx;
+		ws.ws_row = sy;
+		pid = fdforkpty(ptm_fd, &master, NULL, NULL, &ws);
+	} else {
+		if (socketpair(AF_UNIX, SOCK_STREAM, PF_UNSPEC, out) != 0)
+			goto fail;
+		pid = fork();
+	}
+	if (cmd == NULL) {
+		cmd_log_argv(argc, argv, "%s:", __func__);
+		log_debug("%s: cwd=%s", __func__, cwd == NULL ? "" : cwd);
+	} else {
+		log_debug("%s: cmd=%s, cwd=%s", __func__, cmd,
+		    cwd == NULL ? "" : cwd);
+	}
+
+	switch (pid) {
 	case -1:
-		sigprocmask(SIG_SETMASK, &oldset, NULL);
-		environ_free(env);
-		close(out[0]);
-		close(out[1]);
-		return (NULL);
+		if (~flags & JOB_PTY) {
+			close(out[0]);
+			close(out[1]);
+		}
+		goto fail;
 	case 0:
 		proc_clear_signals(server_proc, 1);
 		sigprocmask(SIG_SETMASK, &oldset, NULL);
 
-		if (cwd == NULL || chdir(cwd) != 0) {
-			if ((home = find_home()) == NULL || chdir(home) != 0)
-				chdir("/");
-		}
+		if ((cwd == NULL || chdir(cwd) != 0) &&
+		    ((home = find_home()) == NULL || chdir(home) != 0) &&
+		    chdir("/") != 0)
+			fatal("chdir failed");
 
 		environ_push(env);
 		environ_free(env);
 
-		if (dup2(out[1], STDIN_FILENO) == -1)
-			fatal("dup2 failed");
-		if (dup2(out[1], STDOUT_FILENO) == -1)
-			fatal("dup2 failed");
-		if (out[1] != STDIN_FILENO && out[1] != STDOUT_FILENO)
-			close(out[1]);
-		close(out[0]);
-
-		nullfd = open(_PATH_DEVNULL, O_RDWR, 0);
-		if (nullfd == -1)
-			fatal("open failed");
-		if (dup2(nullfd, STDERR_FILENO) == -1)
-			fatal("dup2 failed");
-		if (nullfd != STDERR_FILENO)
-			close(nullfd);
+		if (~flags & JOB_PTY) {
+			if (dup2(out[1], STDIN_FILENO) == -1)
+				fatal("dup2 failed");
+			if (dup2(out[1], STDOUT_FILENO) == -1)
+				fatal("dup2 failed");
+			if (out[1] != STDIN_FILENO && out[1] != STDOUT_FILENO)
+				close(out[1]);
+			close(out[0]);
 
+			nullfd = open(_PATH_DEVNULL, O_RDWR, 0);
+			if (nullfd == -1)
+				fatal("open failed");
+			if (dup2(nullfd, STDERR_FILENO) == -1)
+				fatal("dup2 failed");
+			if (nullfd != STDERR_FILENO)
+				close(nullfd);
+		}
 		closefrom(STDERR_FILENO + 1);
 
-		execl(_PATH_BSHELL, "sh", "-c", cmd, (char *) NULL);
-		fatal("execl failed");
+		if (cmd != NULL) {
+			execl(_PATH_BSHELL, "sh", "-c", cmd, (char *) NULL);
+			fatal("execl failed");
+		} else {
+			argvp = cmd_copy_argv(argc, argv);
+			execvp(argvp[0], argvp);
+			fatal("execvp failed");
+		}
 	}
 
 	sigprocmask(SIG_SETMASK, &oldset, NULL);
 	environ_free(env);
-	close(out[1]);
 
 	job = xmalloc(sizeof *job);
 	job->state = JOB_RUNNING;
 	job->flags = flags;
 
-	job->cmd = xstrdup(cmd);
+	if (cmd != NULL)
+		job->cmd = xstrdup(cmd);
+	else
+		job->cmd = cmd_stringify_argv(argc, argv);
 	job->pid = pid;
 	job->status = 0;
 
@@ -149,7 +177,11 @@ job_run(const char *cmd, struct session *s, const char *cwd,
 	job->freecb = freecb;
 	job->data = data;
 
-	job->fd = out[0];
+	if (~flags & JOB_PTY) {
+		close(out[1]);
+		job->fd = out[0];
+	} else
+		job->fd = master;
 	setblocking(job->fd, 0);
 
 	job->event = bufferevent_new(job->fd, job_read_callback,
@@ -160,6 +192,11 @@ job_run(const char *cmd, struct session *s, const char *cwd,
 
 	log_debug("run job %p: %s, pid %ld", job, job->cmd, (long) job->pid);
 	return (job);
+
+fail:
+	sigprocmask(SIG_SETMASK, &oldset, NULL);
+	environ_free(env);
+	return (NULL);
 }
 
 /* Kill and free an individual job. */
@@ -184,6 +221,24 @@ job_free(struct job *job)
 	free(job);
 }
 
+/* Resize job. */
+void
+job_resize(struct job *job, u_int sx, u_int sy)
+{
+	struct winsize	 ws;
+
+	if (job->fd == -1 || (~job->flags & JOB_PTY))
+		return;
+
+	log_debug("resize job %p: %ux%u", job, sx, sy);
+
+	memset(&ws, 0, sizeof ws);
+	ws.ws_col = sx;
+	ws.ws_row = sy;
+	if (ioctl(job->fd, TIOCSWINSZ, &ws) == -1)
+		fatal("ioctl failed");
+}
+
 /* Job buffer read callback. */
 static void
 job_read_callback(__unused struct bufferevent *bufev, void *data)
@@ -208,7 +263,7 @@ job_write_callback(__unused struct bufferevent *bufev, void *data)
 	log_debug("job write %p: %s, pid %ld, output left %zu", job, job->cmd,
 	    (long) job->pid, len);
 
-	if (len == 0) {
+	if (len == 0 && (~job->flags & JOB_KEEPWRITE)) {
 		shutdown(job->fd, SHUT_WR);
 		bufferevent_disable(job->event, EV_WRITE);
 	}
@@ -245,6 +300,12 @@ job_check_died(pid_t pid, int status)
 	}
 	if (job == NULL)
 		return;
+	if (WIFSTOPPED(status)) {
+		if (WSTOPSIG(status) == SIGTTIN || WSTOPSIG(status) == SIGTTOU)
+			return;
+		killpg(job->pid, SIGCONT);
+		return;
+	}
 	log_debug("job died %p: %s, pid %ld", job, job->cmd, (long) job->pid);
 
 	job->status = status;
