@@ -30,7 +30,6 @@
 
 #include "tmux.h"
 
-static void	 tty_term_override(struct tty_term *, const char *);
 static char	*tty_term_strip(const char *);
 
 struct tty_terms tty_terms = LIST_HEAD_INITIALIZER(tty_terms);
@@ -58,13 +57,17 @@ struct tty_term_code_entry {
 
 static const struct tty_term_code_entry tty_term_codes[] = {
 	[TTYC_ACSC] = { TTYCODE_STRING, "acsc" },
+	[TTYC_AM] = { TTYCODE_FLAG, "am" },
 	[TTYC_AX] = { TTYCODE_FLAG, "AX" },
 	[TTYC_BCE] = { TTYCODE_FLAG, "bce" },
 	[TTYC_BEL] = { TTYCODE_STRING, "bel" },
+	[TTYC_BIDI] = { TTYCODE_STRING, "Bidi" },
 	[TTYC_BLINK] = { TTYCODE_STRING, "blink" },
 	[TTYC_BOLD] = { TTYCODE_STRING, "bold" },
 	[TTYC_CIVIS] = { TTYCODE_STRING, "civis" },
 	[TTYC_CLEAR] = { TTYCODE_STRING, "clear" },
+	[TTYC_CLMG] = { TTYCODE_STRING, "Clmg" },
+	[TTYC_CMG] = { TTYCODE_STRING, "Cmg" },
 	[TTYC_CNORM] = { TTYCODE_STRING, "cnorm" },
 	[TTYC_COLORS] = { TTYCODE_NUMBER, "colors" },
 	[TTYC_CR] = { TTYCODE_STRING, "Cr" },
@@ -85,12 +88,20 @@ static const struct tty_term_code_entry tty_term_codes[] = {
 	[TTYC_DIM] = { TTYCODE_STRING, "dim" },
 	[TTYC_DL1] = { TTYCODE_STRING, "dl1" },
 	[TTYC_DL] = { TTYCODE_STRING, "dl" },
+	[TTYC_DSEKS] = { TTYCODE_STRING, "Dseks" },
+	[TTYC_DSFCS] = { TTYCODE_STRING, "Dsfcs" },
+	[TTYC_DSBP] = { TTYCODE_STRING, "Dsbp" },
+	[TTYC_DSMG] = { TTYCODE_STRING, "Dsmg" },
 	[TTYC_E3] = { TTYCODE_STRING, "E3" },
 	[TTYC_ECH] = { TTYCODE_STRING, "ech" },
 	[TTYC_ED] = { TTYCODE_STRING, "ed" },
 	[TTYC_EL1] = { TTYCODE_STRING, "el1" },
 	[TTYC_EL] = { TTYCODE_STRING, "el" },
 	[TTYC_ENACS] = { TTYCODE_STRING, "enacs" },
+	[TTYC_ENBP] = { TTYCODE_STRING, "Enbp" },
+	[TTYC_ENEKS] = { TTYCODE_STRING, "Eneks" },
+	[TTYC_ENFCS] = { TTYCODE_STRING, "Enfcs" },
+	[TTYC_ENMG] = { TTYCODE_STRING, "Enmg" },
 	[TTYC_FSL] = { TTYCODE_STRING, "fsl" },
 	[TTYC_HOME] = { TTYCODE_STRING, "home" },
 	[TTYC_HPA] = { TTYCODE_STRING, "hpa" },
@@ -238,16 +249,19 @@ static const struct tty_term_code_entry tty_term_codes[] = {
 	[TTYC_KUP6] = { TTYCODE_STRING, "kUP6" },
 	[TTYC_KUP7] = { TTYCODE_STRING, "kUP7" },
 	[TTYC_MS] = { TTYCODE_STRING, "Ms" },
+	[TTYC_OL] = { TTYCODE_STRING, "ol" },
 	[TTYC_OP] = { TTYCODE_STRING, "op" },
+	[TTYC_RECT] = { TTYCODE_STRING, "Rect" },
 	[TTYC_REV] = { TTYCODE_STRING, "rev" },
 	[TTYC_RGB] = { TTYCODE_FLAG, "RGB" },
-	[TTYC_RI] = { TTYCODE_STRING, "ri" },
 	[TTYC_RIN] = { TTYCODE_STRING, "rin" },
+	[TTYC_RI] = { TTYCODE_STRING, "ri" },
 	[TTYC_RMACS] = { TTYCODE_STRING, "rmacs" },
 	[TTYC_RMCUP] = { TTYCODE_STRING, "rmcup" },
 	[TTYC_RMKX] = { TTYCODE_STRING, "rmkx" },
 	[TTYC_SETAB] = { TTYCODE_STRING, "setab" },
 	[TTYC_SETAF] = { TTYCODE_STRING, "setaf" },
+	[TTYC_SETAL] = { TTYCODE_STRING, "setal" },
 	[TTYC_SETRGBB] = { TTYCODE_STRING, "setrgbb" },
 	[TTYC_SETRGBF] = { TTYCODE_STRING, "setrgbf" },
 	[TTYC_SETULC] = { TTYCODE_STRING, "Setulc" },
@@ -263,12 +277,12 @@ static const struct tty_term_code_entry tty_term_codes[] = {
 	[TTYC_SMUL] = { TTYCODE_STRING, "smul" },
 	[TTYC_SMXX] =  { TTYCODE_STRING, "smxx" },
 	[TTYC_SS] = { TTYCODE_STRING, "Ss" },
+	[TTYC_SYNC] = { TTYCODE_STRING, "Sync" },
 	[TTYC_TC] = { TTYCODE_FLAG, "Tc" },
 	[TTYC_TSL] = { TTYCODE_STRING, "tsl" },
 	[TTYC_U8] = { TTYCODE_NUMBER, "U8" },
 	[TTYC_VPA] = { TTYCODE_STRING, "vpa" },
-	[TTYC_XENL] = { TTYCODE_FLAG, "xenl" },
-	[TTYC_XT] = { TTYCODE_FLAG, "XT" },
+	[TTYC_XT] = { TTYCODE_FLAG, "XT" }
 };
 
 u_int
@@ -295,6 +309,8 @@ tty_term_strip(const char *s)
 				ptr++;
 			if (*ptr == '>')
 				ptr++;
+			if (*ptr == '\0')
+				break;
 		}
 
 		buf[len++] = *ptr;
@@ -337,22 +353,18 @@ tty_term_override_next(const char *s, size_t *offset)
 	return (value);
 }
 
-static void
-tty_term_override(struct tty_term *term, const char *override)
+void
+tty_term_apply(struct tty_term *term, const char *capabilities, int quiet)
 {
 	const struct tty_term_code_entry	*ent;
 	struct tty_code				*code;
 	size_t                                   offset = 0;
 	char					*cp, *value, *s;
-	const char				*errstr;
+	const char				*errstr, *name = term->name;
 	u_int					 i;
 	int					 n, remove;
 
-	s = tty_term_override_next(override, &offset);
-	if (s == NULL || fnmatch(s, term->name, 0) != 0)
-		return;
-
-	while ((s = tty_term_override_next(override, &offset)) != NULL) {
+	while ((s = tty_term_override_next(capabilities, &offset)) != NULL) {
 		if (*s == '\0')
 			continue;
 		value = NULL;
@@ -371,12 +383,14 @@ tty_term_override(struct tty_term *term, const char *override)
 		} else
 			value = xstrdup("");
 
-		if (remove)
-			log_debug("%s override: %s@", term->name, s);
-		else if (*value == '\0')
-			log_debug("%s override: %s", term->name, s);
-		else
-			log_debug("%s override: %s=%s", term->name, s, value);
+		if (!quiet) {
+			if (remove)
+				log_debug("%s override: %s@", name, s);
+			else if (*value == '\0')
+				log_debug("%s override: %s", name, s);
+			else
+				log_debug("%s override: %s=%s", name, s, value);
+		}
 
 		for (i = 0; i < tty_term_ncodes(); i++) {
 			ent = &tty_term_codes[i];
@@ -415,8 +429,92 @@ tty_term_override(struct tty_term *term, const char *override)
 	}
 }
 
+void
+tty_term_apply_overrides(struct tty_term *term)
+{
+	struct options_entry		*o;
+	struct options_array_item	*a;
+	union options_value		*ov;
+	const char			*s, *acs;
+	size_t				 offset;
+	char				*first;
+
+	/* Update capabilities from the option. */
+	o = options_get_only(global_options, "terminal-overrides");
+	a = options_array_first(o);
+	while (a != NULL) {
+		ov = options_array_item_value(a);
+		s = ov->string;
+
+		offset = 0;
+		first = tty_term_override_next(s, &offset);
+		if (first != NULL && fnmatch(first, term->name, 0) == 0)
+			tty_term_apply(term, s + offset, 0);
+		a = options_array_next(a);
+	}
+
+	/* Update the RGB flag if the terminal has RGB colours. */
+	if (tty_term_has(term, TTYC_SETRGBF) &&
+	    tty_term_has(term, TTYC_SETRGBB))
+		term->flags |= TERM_RGBCOLOURS;
+	else
+		term->flags &= ~TERM_RGBCOLOURS;
+	log_debug("RGBCOLOURS flag is %d", !!(term->flags & TERM_RGBCOLOURS));
+
+	/*
+	 * Set or clear the DECSLRM flag if the terminal has the margin
+	 * capabilities.
+	 */
+	if (tty_term_has(term, TTYC_CMG) && tty_term_has(term, TTYC_CLMG))
+		term->flags |= TERM_DECSLRM;
+	else
+		term->flags &= ~TERM_DECSLRM;
+	log_debug("DECSLRM flag is %d", !!(term->flags & TERM_DECSLRM));
+
+	/*
+	 * Set or clear the DECFRA flag if the terminal has the rectangle
+	 * capability.
+	 */
+	if (tty_term_has(term, TTYC_RECT))
+		term->flags |= TERM_DECFRA;
+	else
+		term->flags &= ~TERM_DECFRA;
+	log_debug("DECFRA flag is %d", !!(term->flags & TERM_DECFRA));
+
+	/*
+	 * Terminals without am (auto right margin) wrap at at $COLUMNS - 1
+	 * rather than $COLUMNS (the cursor can never be beyond $COLUMNS - 1).
+	 *
+	 * Terminals without xenl (eat newline glitch) ignore a newline beyond
+	 * the right edge of the terminal, but tmux doesn't care about this -
+	 * it always uses absolute only moves the cursor with a newline when
+	 * also sending a linefeed.
+	 *
+	 * This is irritating, most notably because it is painful to write to
+	 * the very bottom-right of the screen without scrolling.
+	 *
+	 * Flag the terminal here and apply some workarounds in other places to
+	 * do the best possible.
+	 */
+	if (!tty_term_flag(term, TTYC_AM))
+		term->flags |= TERM_NOAM;
+	else
+		term->flags &= ~TERM_NOAM;
+	log_debug("NOAM flag is %d", !!(term->flags & TERM_NOAM));
+
+	/* Generate ACS table. If none is present, use nearest ASCII. */
+	memset(term->acs, 0, sizeof term->acs);
+	if (tty_term_has(term, TTYC_ACSC))
+		acs = tty_term_string(term, TTYC_ACSC);
+	else
+		acs = "a#j+k+l+m+n+o-p-q-r-s-t+u+v+w+x|y<z>~.";
+	for (; acs[0] != '\0' && acs[1] != '\0'; acs += 2)
+		term->acs[(u_char) acs[0]][0] = acs[1];
+}
+
 struct tty_term *
-tty_term_find(char *name, int fd, char **cause)
+tty_term_create(struct tty *tty, char *name, char **caps, u_int ncaps,
+    int *feat, char **cause)
 {
 	struct tty_term				*term;
 	const struct tty_term_code_entry	*ent;
@@ -424,26 +522,154 @@ tty_term_find(char *name, int fd, char **cause)
 	struct options_entry			*o;
 	struct options_array_item		*a;
 	union options_value			*ov;
-	u_int					 i;
-	int		 			 n, error;
-	const char				*s, *acs;
+	u_int					 i, j;
+	const char				*s, *value;
+	size_t					 offset, namelen;
+	char					*first;
 
-	LIST_FOREACH(term, &tty_terms, entry) {
-		if (strcmp(term->name, name) == 0) {
-			term->references++;
-			return (term);
-		}
-	}
-	log_debug("new term: %s", name);
+	log_debug("adding term %s", name);
 
-	term = xmalloc(sizeof *term);
+	term = xcalloc(1, sizeof *term);
+	term->tty = tty;
 	term->name = xstrdup(name);
-	term->references = 1;
-	term->flags = 0;
 	term->codes = xcalloc(tty_term_ncodes(), sizeof *term->codes);
 	LIST_INSERT_HEAD(&tty_terms, term, entry);
 
-	/* Set up curses terminal. */
+	/* Fill in codes. */
+	for (i = 0; i < ncaps; i++) {
+		namelen = strcspn(caps[i], "=");
+		if (namelen == 0)
+			continue;
+		value = caps[i] + namelen + 1;
+
+		for (j = 0; j < tty_term_ncodes(); j++) {
+			ent = &tty_term_codes[j];
+			if (strncmp(ent->name, caps[i], namelen) != 0)
+				continue;
+			if (ent->name[namelen] != '\0')
+				continue;
+
+			code = &term->codes[j];
+			code->type = TTYCODE_NONE;
+			switch (ent->type) {
+			case TTYCODE_NONE:
+				break;
+			case TTYCODE_STRING:
+				code->type = TTYCODE_STRING;
+				code->value.string = tty_term_strip(value);
+				break;
+			case TTYCODE_NUMBER:
+				code->type = TTYCODE_NUMBER;
+				code->value.number = atoi(value);
+				break;
+			case TTYCODE_FLAG:
+				code->type = TTYCODE_FLAG;
+				code->value.flag = (*value == '1');
+				break;
+			}
+		}
+	}
+
+	/* Apply terminal features. */
+	o = options_get_only(global_options, "terminal-features");
+	a = options_array_first(o);
+	while (a != NULL) {
+		ov = options_array_item_value(a);
+		s = ov->string;
+
+		offset = 0;
+		first = tty_term_override_next(s, &offset);
+		if (first != NULL && fnmatch(first, term->name, 0) == 0)
+			tty_add_features(feat, s + offset, ":");
+		a = options_array_next(a);
+	}
+
+	/* Delete curses data. */
+#if !defined(NCURSES_VERSION_MAJOR) || NCURSES_VERSION_MAJOR > 5 || \
+    (NCURSES_VERSION_MAJOR == 5 && NCURSES_VERSION_MINOR > 6)
+	del_curterm(cur_term);
+#endif
+
+	/* Apply overrides so any capabilities used for features are changed. */
+	tty_term_apply_overrides(term);
+
+	/* These are always required. */
+	if (!tty_term_has(term, TTYC_CLEAR)) {
+		xasprintf(cause, "terminal does not support clear");
+		goto error;
+	}
+	if (!tty_term_has(term, TTYC_CUP)) {
+		xasprintf(cause, "terminal does not support cup");
+		goto error;
+	}
+
+	/*
+	 * If TERM has XT or clear starts with CSI then it is safe to assume
+	 * the terminal is derived from the VT100. This controls whether device
+	 * attributes requests are sent to get more information.
+	 *
+	 * This is a bit of a hack but there aren't that many alternatives.
+	 * Worst case tmux will just fall back to using whatever terminfo(5)
+	 * says without trying to correct anything that is missing.
+	 *
+	 * Also add few features that VT100-like terminals should either
+	 * support or safely ignore.
+	 */
+	s = tty_term_string(term, TTYC_CLEAR);
+	if (tty_term_flag(term, TTYC_XT) || strncmp(s, "\033[", 2) == 0) {
+		term->flags |= TERM_VT100LIKE;
+		tty_add_features(feat, "bpaste,focus,title", ",");
+	}
+
+	/* Add RGB feature if terminal has RGB colours. */
+	if ((tty_term_flag(term, TTYC_TC) || tty_term_has(term, TTYC_RGB)) &&
+	    (!tty_term_has(term, TTYC_SETRGBF) ||
+	    !tty_term_has(term, TTYC_SETRGBB)))
+		tty_add_features(feat, "RGB", ",");
+
+	/* Apply the features and overrides again. */
+	if (tty_apply_features(term, *feat))
+		tty_term_apply_overrides(term);
+
+	/* Log the capabilities. */
+	for (i = 0; i < tty_term_ncodes(); i++)
+		log_debug("%s%s", name, tty_term_describe(term, i));
+
+	return (term);
+
+error:
+	tty_term_free(term);
+	return (NULL);
+}
+
+void
+tty_term_free(struct tty_term *term)
+{
+	u_int	i;
+
+	log_debug("removing term %s", term->name);
+
+	for (i = 0; i < tty_term_ncodes(); i++) {
+		if (term->codes[i].type == TTYCODE_STRING)
+			free(term->codes[i].value.string);
+	}
+	free(term->codes);
+
+	LIST_REMOVE(term, entry);
+	free(term->name);
+	free(term);
+}
+
+int
+tty_term_read_list(const char *name, int fd, char ***caps, u_int *ncaps,
+    char **cause)
+{
+	const struct tty_term_code_entry	*ent;
+	int					 error, n;
+	u_int					 i;
+	const char				*s;
+	char					 tmp[11];
+
 	if (setupterm(name, fd, &error) != OK) {
 		switch (error) {
 		case 1:
@@ -461,156 +687,59 @@ tty_term_find(char *name, int fd, char **cause)
 			xasprintf(cause, "unknown error");
 			break;
 		}
-		goto error;
+		return (-1);
 	}
 
-	/* Fill in codes. */
+	*ncaps = 0;
+	*caps = NULL;
+
 	for (i = 0; i < tty_term_ncodes(); i++) {
 		ent = &tty_term_codes[i];
-
-		code = &term->codes[i];
-		code->type = TTYCODE_NONE;
 		switch (ent->type) {
 		case TTYCODE_NONE:
-			break;
+			continue;
 		case TTYCODE_STRING:
-			s = tigetstr((char *) ent->name);
-			if (s == NULL || s == (char *) -1)
-				break;
-			code->type = TTYCODE_STRING;
-			code->value.string = tty_term_strip(s);
+			s = tigetstr((char *)ent->name);
+			if (s == NULL || s == (char *)-1)
+				continue;
 			break;
 		case TTYCODE_NUMBER:
-			n = tigetnum((char *) ent->name);
+			n = tigetnum((char *)ent->name);
 			if (n == -1 || n == -2)
-				break;
-			code->type = TTYCODE_NUMBER;
-			code->value.number = n;
+				continue;
+			xsnprintf(tmp, sizeof tmp, "%d", n);
+			s = tmp;
 			break;
 		case TTYCODE_FLAG:
 			n = tigetflag((char *) ent->name);
 			if (n == -1)
-				break;
-			code->type = TTYCODE_FLAG;
-			code->value.flag = n;
+				continue;
+			if (n)
+				s = "1";
+			else
+				s = "0";
 			break;
 		}
+		*caps = xreallocarray(*caps, (*ncaps) + 1, sizeof **caps);
+		xasprintf(&(*caps)[*ncaps], "%s=%s", ent->name, s);
+		(*ncaps)++;
 	}
 
-	/* Apply terminal overrides. */
-	o = options_get_only(global_options, "terminal-overrides");
-	a = options_array_first(o);
-	while (a != NULL) {
-		ov = options_array_item_value(a);
-		tty_term_override(term, ov->string);
-		a = options_array_next(a);
-	}
-
-	/* Delete curses data. */
 #if !defined(NCURSES_VERSION_MAJOR) || NCURSES_VERSION_MAJOR > 5 || \
     (NCURSES_VERSION_MAJOR == 5 && NCURSES_VERSION_MINOR > 6)
 	del_curterm(cur_term);
 #endif
-
-	/* These are always required. */
-	if (!tty_term_has(term, TTYC_CLEAR)) {
-		xasprintf(cause, "terminal does not support clear");
-		goto error;
-	}
-	if (!tty_term_has(term, TTYC_CUP)) {
-		xasprintf(cause, "terminal does not support cup");
-		goto error;
-	}
-
-	/* These can be emulated so one of the two is required. */
-	if (!tty_term_has(term, TTYC_CUD1) && !tty_term_has(term, TTYC_CUD)) {
-		xasprintf(cause, "terminal does not support cud1 or cud");
-		goto error;
-	}
-
-	/* Figure out if we have 256 colours (or more). */
-	if (tty_term_number(term, TTYC_COLORS) >= 256 ||
-	    tty_term_has(term, TTYC_RGB))
-		term->flags |= TERM_256COLOURS;
-
-	/*
-	 * Terminals without xenl (eat newline glitch) wrap at at $COLUMNS - 1
-	 * rather than $COLUMNS (the cursor can never be beyond $COLUMNS - 1).
-	 *
-	 * This is irritating, most notably because it is impossible to write
-	 * to the very bottom-right of the screen without scrolling.
-	 *
-	 * Flag the terminal here and apply some workarounds in other places to
-	 * do the best possible.
-	 */
-	if (!tty_term_flag(term, TTYC_XENL))
-		term->flags |= TERM_EARLYWRAP;
-
-	/* Generate ACS table. If none is present, use nearest ASCII. */
-	memset(term->acs, 0, sizeof term->acs);
-	if (tty_term_has(term, TTYC_ACSC))
-		acs = tty_term_string(term, TTYC_ACSC);
-	else
-		acs = "a#j+k+l+m+n+o-p-q-r-s-t+u+v+w+x|y<z>~.";
-	for (; acs[0] != '\0' && acs[1] != '\0'; acs += 2)
-		term->acs[(u_char) acs[0]][0] = acs[1];
-
-	/* On terminals with xterm titles (XT), fill in tsl and fsl. */
-	if (tty_term_flag(term, TTYC_XT) &&
-	    !tty_term_has(term, TTYC_TSL) &&
-	    !tty_term_has(term, TTYC_FSL)) {
-		code = &term->codes[TTYC_TSL];
-		code->value.string = xstrdup("\033]0;");
-		code->type = TTYCODE_STRING;
-		code = &term->codes[TTYC_FSL];
-		code->value.string = xstrdup("\007");
-		code->type = TTYCODE_STRING;
-	}
-
-	/*
-	 * On terminals with RGB colour (Tc or RGB), fill in setrgbf and
-	 * setrgbb if they are missing.
-	 */
-	if ((tty_term_flag(term, TTYC_TC) || tty_term_flag(term, TTYC_RGB)) &&
-	    !tty_term_has(term, TTYC_SETRGBF) &&
-	    !tty_term_has(term, TTYC_SETRGBB)) {
-		code = &term->codes[TTYC_SETRGBF];
-		code->value.string = xstrdup("\033[38;2;%p1%d;%p2%d;%p3%dm");
-		code->type = TTYCODE_STRING;
-		code = &term->codes[TTYC_SETRGBB];
-		code->value.string = xstrdup("\033[48;2;%p1%d;%p2%d;%p3%dm");
-		code->type = TTYCODE_STRING;
-	}
-
-	/* Log it. */
-	for (i = 0; i < tty_term_ncodes(); i++)
-		log_debug("%s%s", name, tty_term_describe(term, i));
-
-	return (term);
-
-error:
-	tty_term_free(term);
-	return (NULL);
+	return (0);
 }
 
 void
-tty_term_free(struct tty_term *term)
+tty_term_free_list(char **caps, u_int ncaps)
 {
 	u_int	i;
 
-	if (--term->references != 0)
-		return;
-
-	LIST_REMOVE(term, entry);
-
-	for (i = 0; i < tty_term_ncodes(); i++) {
-		if (term->codes[i].type == TTYCODE_STRING)
-			free(term->codes[i].value.string);
-	}
-	free(term->codes);
-
-	free(term->name);
-	free(term);
+	for (i = 0; i < ncaps; i++)
+		free(caps[i]);
+	free(caps);
 }
 
 int
@@ -642,7 +771,8 @@ tty_term_string2(struct tty_term *term, enum tty_code_code code, int a, int b)
 }
 
 const char *
-tty_term_string3(struct tty_term *term, enum tty_code_code code, int a, int b, int c)
+tty_term_string3(struct tty_term *term, enum tty_code_code code, int a, int b,
+    int c)
 {
 	return (tparm((char *) tty_term_string(term, code), a, b, c, 0, 0, 0, 0, 0, 0));
 }
